﻿using ArcSoftFace.Utils;
using System;
using System.Drawing;
using System.IO;
using System.Net;
using System.Text;

namespace ArcSoftFace.http
{
    /// <summary>
    /// Http请求帮助类
    /// </summary>
    public class HttpHelper
    {

        /// <summary>
        /// Post提交数据
        /// </summary>
        /// <param name="url">url地址</param>
        /// <param name="body">参数</param>
        /// <param name="contenttype">参数</param>
        /// <returns></returns>
        public static string HttpPost(string url, string body, string contenttype = "")
        {
            try
            {
                Encoding encoding = Encoding.UTF8;
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url);
                request.Method = "POST";
                request.Accept = "application/json, text/javascript, */*"; //"text/html, application/xhtml+xml, */*";
                request.ContentType = "application/json; charset=utf-8";
                if (contenttype != "")
                {
                    request.ContentType = contenttype;
                }
                byte[] buffer = encoding.GetBytes(body);
                request.ContentLength = buffer.Length;
                request.GetRequestStream().Write(buffer, 0, buffer.Length);
                HttpWebResponse response = (HttpWebResponse)request.GetResponse();
                using StreamReader reader = new StreamReader(response.GetResponseStream(), encoding);
                return reader.ReadToEnd();
            }
            catch (WebException ex)
            {
                //var res = (HttpWebResponse)ex.Response;
                //StringBuilder sb = new StringBuilder();
                //StreamReader sr = new StreamReader(res.GetResponseStream(), Encoding.UTF8);
                //sb.Append(sr.ReadToEnd());
                //throw new Exception(sb.ToString());
                return "";
            }
        }

        /// <summary>
        /// Get获取数据
        /// </summary>
        /// <param name="url">url地址</param>
        /// <param name="encode">编码方式</param>
        /// <param name="Source">来源</param>
        /// <returns></returns>
        public static string HttpGet(string url, Encoding encode, string Source, string cookie="")
        {
            HttpWebRequest myRequest = (HttpWebRequest)WebRequest.Create(url);
            if (!string.IsNullOrEmpty(Source))
            {
                myRequest.Referer = Source;
            }
            if (!string.IsNullOrEmpty(cookie))
            {
                myRequest.Headers.Add("cookie", cookie);
            }
            myRequest.Method = "GET";
            HttpWebResponse myResponse;
            try
            {
                myResponse = (HttpWebResponse)myRequest.GetResponse();
                StreamReader reader = new StreamReader(myResponse.GetResponseStream(), encode);
                string content = reader.ReadToEnd();
                return content;
            }
            //异常请求  
            catch (WebException e)
            {
                myResponse = (HttpWebResponse)e.Response;
                using Stream errData = myResponse.GetResponseStream();
                using StreamReader reader = new StreamReader(errData);
                string text = reader.ReadToEnd();
                return text;
            }
        }

        /// <summary>
        /// 执行HTTP GET请求。
        /// </summary>
        /// <param name="url">请求地址</param>
        /// <returns>HTTP响应</returns>
        public static string HttpGet(string url)
        {
            HttpWebRequest req = (HttpWebRequest)WebRequest.Create(url);
            req.Method = "GET";
            req.ContentType = "application/json";
            HttpWebResponse rsp = (HttpWebResponse)req.GetResponse();
            if (string.IsNullOrWhiteSpace(rsp.CharacterSet))
            {
                return GetResponseAsString(rsp, Encoding.UTF8);
            }
            else
            {
                Encoding encoding = Encoding.GetEncoding(rsp.CharacterSet);
                return GetResponseAsString(rsp, encoding);
            }
        }


        /// <summary>
        /// 执行HTTP GET请求。
        /// </summary>
        /// <param name="url">请求地址</param>
        /// <returns>HTTP响应</returns>
        public static string HttpPostStr(string url, string postDataStr)
        {
            HttpWebRequest req = (HttpWebRequest)WebRequest.Create(url);
            req.Method = "POST";
            req.ContentType = "application/x-www-form-urlencoded";
            byte[] data = Encoding.UTF8.GetBytes(postDataStr);
            req.ContentLength = data.Length;
            using (Stream reqStream = req.GetRequestStream())
            {
                reqStream.Write(data, 0, data.Length);
                reqStream.Close();
            }
            HttpWebResponse rsp = (HttpWebResponse)req.GetResponse();
            if (string.IsNullOrWhiteSpace(rsp.CharacterSet))
            {
                return GetResponseAsString(rsp, Encoding.UTF8);
            }
            else
            {
                Encoding encoding = Encoding.GetEncoding(rsp.CharacterSet);
                return GetResponseAsString(rsp, encoding);
            }
        }

        /// <summary>
        /// 执行HTTP POST请求。
        /// </summary>
        /// <param name="url">请求地址</param>        
        /// <param name="postDataStr">请求地址</param>        
        /// <returns>HTTP响应</returns>
        
        /// <summary>
        /// 把响应流转换为文本。
        /// </summary>
        /// <param name="rsp">响应流对象</param>
        /// <param name="encoding">编码方式</param>
        /// <returns>响应文本</returns>
        private static string GetResponseAsString(HttpWebResponse rsp, Encoding encoding)
        {
            System.IO.Stream stream = null;
            StreamReader reader = null;
            try
            {
                // 以字符流的方式读取HTTP响应
                stream = rsp.GetResponseStream();
                reader = new StreamReader(stream, encoding);
                return reader.ReadToEnd();
            }
            finally
            {
                // 释放资源
                if (reader != null) reader.Close();
                if (stream != null) stream.Close();
                if (rsp != null) rsp.Close();
            }
        }

        public Bitmap UrlToImage(string url, bool IsCheckKongBai)
        {
            try
            {
                WebClient mywebclient = new WebClient();



                url = url.Substring(0, 5).Contains("http") ? url : "http:" + url;

                byte[] Bytes = mywebclient.DownloadData(url);

                using (MemoryStream ms = new MemoryStream(Bytes))
                {
                    Image outputImg = Image.FromStream(ms);

                    if (IsCheckKongBai)
                    {
                        bool IsKongBai = CheckTransparentImg(outputImg);
                        if (IsKongBai)
                        {
                            Console.WriteLine("空白图----" + url);
                            return null;
                        }
                    }

                    //判断是否是png的图
                    if (url.IndexOf(".png")!=-1)
                    {
                        Bitmap bmp = new Bitmap(outputImg.Width, outputImg.Height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);

                        using (Graphics g = Graphics.FromImage(bmp))
                        {
                            g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
                            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
                            g.CompositingQuality = System.Drawing.Drawing2D.CompositingQuality.HighQuality;
                            g.DrawImage(outputImg, 0, 0);
                            g.Dispose();
                        }

                        return bmp;


                    }
                    else
                    {
                        Graphics g = Graphics.FromImage(outputImg);
                        ms.Close();
                        Bitmap b = new Bitmap(outputImg);
                        return b;
                    }


                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString() + url + Environment.NewLine);
                return null;
            }
        }


        /// <summary>
        /// 检查图片是否空白图片
        /// </summary>
        /// <param name="img">Image</param>
        /// <returns>是否空白图片</returns>
        private static bool CheckTransparentImg(Image img)
        {
            bool blnIsTransparent = false;
            //加载位图
            Bitmap bitMap = new Bitmap(img);
            //图片总像素
            int intAll = img.Height * img.Width;
            int intCount = 0;
            //按像素遍历
            for (int intY = 0; intY < img.Height; intY++)
            {
                for (int intX = 0; intX < img.Width; intX++)
                {
                    if (bitMap.GetPixel(intX, intY).GetBrightness() == 1)
                    {
                        intCount += 1;
                    }
                }
            }

            float a = (float)intCount / (float)intAll;
            Console.WriteLine(a.ToString());
            Console.WriteLine(intCount.ToString() + "---" + intAll.ToString());


            //释放资源
            bitMap.Dispose();
            int al = intAll - 1000;
            if (intCount > al)
            {
                blnIsTransparent = true;
            }
            return blnIsTransparent;
        }
    }
}